"""Instrumentation column additions

Revision ID: d5d88ac1d291
Revises: 1bd2f7711e68
Create Date: 2019-10-29 16:51:08.635880

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd5d88ac1d291'
down_revision = '1bd2f7711e68'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('server_stats_log',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('collection_count', sa.Integer(), nullable=True),
    sa.Column('molecule_count', sa.Integer(), nullable=True),
    sa.Column('result_count', sa.Integer(), nullable=True),
    sa.Column('kvstore_count', sa.Integer(), nullable=True),
    sa.Column('access_count', sa.Integer(), nullable=True),
    sa.Column('result_states', sa.JSON(), nullable=True),
    sa.Column('db_total_size', sa.BigInteger(), nullable=True),
    sa.Column('db_table_size', sa.BigInteger(), nullable=True),
    sa.Column('db_index_size', sa.BigInteger(), nullable=True),
    sa.Column('db_table_information', sa.JSON(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('ix_server_stats_log_timestamp', 'server_stats_log', ['timestamp'], unique=False)
    op.create_table('queue_manager_logs',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('manager_id', sa.Integer(), nullable=False),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('completed', sa.Integer(), nullable=True),
    sa.Column('submitted', sa.Integer(), nullable=True),
    sa.Column('failures', sa.Integer(), nullable=True),
    sa.Column('total_worker_walltime', sa.Float(), nullable=True),
    sa.Column('total_task_walltime', sa.Float(), nullable=True),
    sa.Column('active_tasks', sa.Integer(), nullable=True),
    sa.Column('active_cores', sa.Integer(), nullable=True),
    sa.Column('active_memory', sa.Float(), nullable=True),
    sa.ForeignKeyConstraint(['manager_id'], ['queue_manager.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('ix_queue_manager_log_timestamp', 'queue_manager_logs', ['timestamp'], unique=False)
    op.add_column('queue_manager', sa.Column('active_cores', sa.Integer(), nullable=True))
    op.add_column('queue_manager', sa.Column('active_memory', sa.Float(), nullable=True))
    op.add_column('queue_manager', sa.Column('active_tasks', sa.Integer(), nullable=True))
    op.add_column('queue_manager', sa.Column('configuration', sa.JSON(), nullable=True))
    op.add_column('queue_manager', sa.Column('total_task_walltime', sa.Float(), nullable=True))
    op.add_column('queue_manager', sa.Column('total_worker_walltime', sa.Float(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('queue_manager', 'total_worker_walltime')
    op.drop_column('queue_manager', 'total_task_walltime')
    op.drop_column('queue_manager', 'configuration')
    op.drop_column('queue_manager', 'active_tasks')
    op.drop_column('queue_manager', 'active_memory')
    op.drop_column('queue_manager', 'active_cores')
    op.drop_index('ix_queue_manager_log_timestamp', table_name='queue_manager_logs')
    op.drop_table('queue_manager_logs')
    op.drop_index('ix_server_stats_log_timestamp', table_name='server_stats_log')
    op.drop_table('server_stats_log')
    # ### end Alembic commands ###



"""Add contributed values table

Revision ID: 159ba85908fd
Revises: d5d88ac1d291
Create Date: 2019-11-01 15:39:50.970246

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.dialects import postgresql
from qcfractal.storage_sockets.models.sql_base import MsgpackExt
from qcfractal.storage_sockets.models.collections_models import ContributedValuesORM

import numpy as np

# revision identifiers, used by Alembic.
revision = '159ba85908fd'
down_revision = 'd5d88ac1d291'
branch_labels = None
depends_on = None


def migrate_contributed_values_data():

    bind = op.get_bind()
    session = orm.Session(bind=bind)

    # Dataset and reaction datasets tables
    ds_ids_data = session.execute('select id, contributed_values_data from dataset;').fetchall()
    print(f"Migrating datasets with ids: {[ds[0] for ds in ds_ids_data]}")

    rds_ids_data = session.execute('select id, contributed_values_data from reaction_dataset;').fetchall()
    print(f"Migrating reaction datasets with ids: {[ds[0] for ds in rds_ids_data]}")

    ds_ids_data.extend(rds_ids_data)

    for ds in ds_ids_data:
        (ds_id, ds_contrib) = ds
        if ds_contrib is None:
            continue

        for key, dict_values in ds_contrib.items():

            idx, vals = [], []
            for key, value in dict_values['values'].items():
                idx.append(key)
                vals.append(value)

            dict_values['values'] = np.array(vals)
            dict_values['index'] = np.array(idx)

            cv = ContributedValuesORM(**dict_values)
            cv.collection_id = ds_id

            session.add(cv)

    session.commit()


def upgrade():

    # rename old column with data
    op.alter_column('dataset', 'contributed_values', new_column_name='contributed_values_data')
    op.alter_column('reaction_dataset', 'contributed_values', new_column_name='contributed_values_data')

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('contributed_values',
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('collection_id', sa.Integer(), nullable=False),
        sa.Column('citations', sa.JSON(), nullable=True),
        sa.Column('theory_level', sa.JSON(), nullable=False),
        sa.Column('theory_level_details', sa.JSON(), nullable=True),
        sa.Column('comments', sa.String(), nullable=True),
        sa.Column('values', MsgpackExt(), nullable=False),
        sa.Column('index', MsgpackExt(), nullable=False),
        sa.Column('external_url', sa.String(), nullable=True),
        sa.Column('doi', sa.String(), nullable=True),
        sa.Column('units', sa.String(), nullable=False),
        sa.Column('values_structure', sa.JSON(), nullable=True, default=lambda: {}),
        sa.ForeignKeyConstraint(['collection_id'], ['collection.id'], ondelete='cascade'),
        sa.PrimaryKeyConstraint('name', 'collection_id')
    )

    op.alter_column('contributed_values', 'values_structure', server_default=None, nullable=False)

    migrate_contributed_values_data()

    op.drop_column('dataset', 'contributed_values_data')
    op.drop_column('reaction_dataset', 'contributed_values_data')


def downgrade():
    # ### Won't work on production data because data will be lost ###

    op.add_column('reaction_dataset', sa.Column('contributed_values', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True))
    op.add_column('dataset', sa.Column('contributed_values', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True))
    op.drop_table('contributed_values')



"""Updates collections table with owner and visibility

Revision ID: a2f76bb7be65
Revises: c194a8ef6acf
Create Date: 2019-10-08 11:06:15.438742

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'a2f76bb7be65'
down_revision = 'c194a8ef6acf'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('collection', sa.Column('description', sa.String(), nullable=True))

    op.add_column('collection', sa.Column('group', sa.String(), nullable=False, server_default='default'))
    op.alter_column('collection', 'group', server_default=None)

    op.add_column('collection', sa.Column('view_url_hdf5', sa.String(), nullable=True))
    op.add_column('collection', sa.Column('view_url_plaintext', sa.String(), nullable=True))
    op.add_column('collection', sa.Column('view_metadata', sa.JSON(), nullable=True))

    op.add_column('collection', sa.Column('view_available', sa.Boolean(), nullable=True))
    op.execute("UPDATE collection SET view_available=false")
    op.alter_column('collection', 'view_available', nullable=False)

    op.add_column('collection', sa.Column('visibility', sa.Boolean(), nullable=True))
    op.execute("UPDATE collection SET visibility=true")
    op.alter_column('collection', 'visibility', nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('collection', 'visibility')
    op.drop_column('collection', 'view_url_hdf5')
    op.drop_column('collection', 'view_url_plaintext')
    op.drop_column('collection', 'view_metadata')
    op.drop_column('collection', 'view_available')
    op.drop_column('collection', 'group')
    op.drop_column('collection', 'metadata')
    op.drop_column('collection', 'description')
    # ### end Alembic commands ###



"""
Common models for QCPortal/Fractal
"""
from enum import Enum
from typing import Any, Dict, Optional

from pydantic import Field, validator
from qcelemental.models import AutodocBaseSettings, Molecule, ProtoModel, Provenance
from qcelemental.models.procedures import OptimizationProtocols
from qcelemental.models.results import ResultProtocols

from .model_utils import hash_dictionary, prepare_basis, recursive_normalizer

__all__ = ["QCSpecification", "OptimizationSpecification", "KeywordSet", "ObjectId", "DriverEnum", "Citation"]

# Add in QCElemental models
__all__.extend(["Molecule", "Provenance", "ProtoModel"])

# Autodoc
__all__.extend(["OptimizationProtocols", "ResultProtocols"])


class ObjectId(str):
    """
    The Id of the object in the data.
    """
    _valid_hex = set("0123456789abcdef")

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if (isinstance(v, str) and (len(v) == 24) and (set(v) <= cls._valid_hex)):
            return v
        elif isinstance(v, int):
            return str(v)
        elif isinstance(v, str) and v.isdigit():
            return (v)
        else:
            raise TypeError("The string {} is not a valid 24-character hexadecimal or integer ObjectId!".format(v))


class DriverEnum(str, Enum):
    """
    The type of calculation that is being performed (e.g., energy, gradient, Hessian, ...).
    """
    energy = 'energy'
    gradient = 'gradient'
    hessian = 'hessian'
    properties = 'properties'


class QCSpecification(ProtoModel):
    """
    The quantum chemistry metadata specification for individual computations such as energy, gradient, and Hessians.
    """
    driver: DriverEnum = Field(..., description=str(DriverEnum.__doc__))
    method: str = Field(..., description="The quantum chemistry method to evaluate (e.g., B3LYP, PBE, ...).")
    basis: Optional[str] = Field(
        None,
        description="The quantum chemistry basis set to evaluate (e.g., 6-31g, cc-pVDZ, ...). Can be ``None`` for "
        "methods without basis sets.")
    keywords: Optional[ObjectId] = Field(
        None,
        description="The Id of the :class:`KeywordSet` registered in the database to run this calculation with. This "
        "Id must exist in the database.")
    protocols: ResultProtocols = Field(ResultProtocols(), description=str(ResultProtocols.__base_doc__))
    program: str = Field(
        ...,
        description="The quantum chemistry program to evaluate the computation with. Not all quantum chemistry programs"
        " support all combinations of driver/method/basis.")

    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)

        # Maintain hash compatability
        if len(ret["protocols"]) == 0:
            ret.pop("protocols", None)

        return ret

    @validator('basis')
    def _check_basis(cls, v):
        return prepare_basis(v)

    @validator('program')
    def _check_program(cls, v):
        return v.lower()

    @validator('method')
    def _check_method(cls, v):
        return v.lower()

    def form_schema_object(self, keywords: Optional['KeywordSet'] = None, checks=True) -> Dict[str, Any]:
        if checks and self.keywords:
            assert keywords.id == self.keywords

        ret = {
            "driver": str(self.driver.name),
            "program": self.program,
            "model": {
                "method": self.method
            }
        } # yapf: disable
        if self.basis:
            ret["model"]["basis"] = self.basis

        if keywords:
            ret["keywords"] = keywords.values
        else:
            ret["keywords"] = {}

        return ret


class OptimizationSpecification(ProtoModel):
    """
    Metadata describing a geometry optimization.
    """
    program: str = Field(..., description="Optimization program to run the optimization with")
    keywords: Optional[Dict[str, Any]] = Field(
        None,
        description="Dictionary of keyword arguments to pass into the ``program`` when the program runs. "
        "Note that unlike :class:`QCSpecification` this is a dictionary of keywords, not the Id for a "
        ":class:`KeywordSet`. ")
    protocols: OptimizationProtocols = Field(OptimizationProtocols(),
                                             description=str(OptimizationProtocols.__base_doc__))

    def dict(self, *args, **kwargs):
        ret = super().dict(*args, **kwargs)

        # Maintain hash compatability
        if len(ret["protocols"]) == 0:
            ret.pop("protocols", None)

        return ret

    @validator('program')
    def _check_program(cls, v):
        return v.lower()

    @validator('keywords')
    def _check_keywords(cls, v):
        if v is not None:
            v = recursive_normalizer(v)
        return v


class KeywordSet(ProtoModel):
    """
    A key:value storage object for Keywords.
    """
    id: Optional[ObjectId] = Field(
        None, description="The Id of this object, will be automatically assigned when added to the database.")
    hash_index: str = Field(
        ...,
        description="The hash of this keyword set to store and check for collisions. This string is automatically "
        "computed.")
    values: Dict[str, Optional[Any]] = Field(
        ...,
        description="The key-value pairs which make up this KeywordSet. There is no direct relation between this "
        "dictionary and applicable program/spec it can be used on.")
    lowercase: bool = Field(
        True,
        description="String keys are in the ``values`` dict are normalized to lowercase if this is True. Assists in "
        "matching against other :class:`KeywordSet` objects in the database.")
    exact_floats: bool = Field(False,
                               description="All floating point numbers are rounded to 1.e-10 if this is False."
                               "Assists in matching against other :class:`KeywordSet` objects in the database.")
    comments: Optional[str] = Field(
        None,
        description="Additional comments for this KeywordSet. Intended for pure human/user consumption "
        "and clarity.")

    def __init__(self, **data):

        build_index = False
        if ("hash_index" not in data) or data.pop("build_index", False):
            build_index = True
            data["hash_index"] = "placeholder"

        ProtoModel.__init__(self, **data)

        # Overwrite options with massaged values
        kwargs = {"lowercase": self.lowercase}
        if self.exact_floats:
            kwargs["digits"] = False

        self.__dict__["values"] = recursive_normalizer(self.values, **kwargs)

        # Build a hash index if we need it
        if build_index:
            self.__dict__["hash_index"] = self.get_hash_index()

    def get_hash_index(self):
        return hash_dictionary(self.values.copy())


class Citation(ProtoModel):
    """ A literature citation.  """
    _acs_citation: str  # hand-formatted citation in ACS style. In the future, this could be bibtex, rendered to different formats.
    _bibtex: Optional[Dict[str, str]]  # bibtex blob for later use with bibtex-renderer
    doi: Optional[str]
    url: Optional[str]

    def to_acs(self) -> str:
        """ Returns an ACS-formatted citation """
        return self._acs_citation



"""Wavefunction data table

Revision ID: 1bd2f7711e68
Revises: a2f76bb7be65
Create Date: 2019-10-17 14:03:07.364322

"""
from alembic import op
import sqlalchemy as sa
import qcfractal


# revision identifiers, used by Alembic.
revision = '1bd2f7711e68'
down_revision = 'a2f76bb7be65'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('wavefunction_store',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('basis', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=False),
    sa.Column('restricted', sa.Boolean(), nullable=False),
    sa.Column('h_core_a', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('h_core_b', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('h_effective_a', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('h_effective_b', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_orbitals_a', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_orbitals_b', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_density_a', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_density_b', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_fock_a', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_fock_b', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_eigenvalues_a', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_eigenvalues_b', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_occupations_a', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('scf_occupations_b', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.Column('extras', qcfractal.storage_sockets.models.sql_base.MsgpackExt(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.add_column('result', sa.Column('wavefunction_data_id', sa.Integer(), nullable=True))
    op.create_foreign_key('result_wavefunction_data_id_fkey', 'result', 'wavefunction_store', ['wavefunction_data_id'], ['id'])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('result_wavefunction_data_id_fkey', 'result', type_='foreignkey')
    op.drop_column('result', 'wavefunction_data_id')
    op.drop_table('wavefunction_store')
    # ### end Alembic commands ###



import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
import pandas as pd
import plotly.graph_objs as go

import dash_coreui_components as coreui
from flask import current_app

from ..connection import get_socket
from ..app import app
from ..dash_models import manager_graph, task_graph, list_managers

## Layout

top_row = dbc.Row(
    [
        dbc.Col(
            coreui.AppCard([
                dbc.CardHeader("Server Information"),
                dbc.CardBody(id='fractal-server-information'),
            ]), ),
        dbc.Col(
            coreui.AppCard([
                dbc.CardHeader("Database Information"),
                dbc.CardBody(id='fractal-database-information'),
            ], ),
            # className="col-sm-6 col-md-3",
        ),
        dbc.Col(
            coreui.AppCard([
                dbc.CardHeader("Queue Information"),
                dbc.CardBody(id='fractal-queue-information'),
            ], ), ),
    ],
    className="w-100")


@app.callback([
    Output('fractal-server-information', 'children'),
    Output('fractal-database-information', 'children'),
    Output('fractal-queue-information', 'children'),
], [Input('manager-overview-groupby', 'value')])
def update_server_information(status):
    config = current_app.config["FRACTAL_CONFIG"]
    server = dcc.Markdown(f"""
**Name:** {config.fractal.name}

**Query Limit:** {config.fractal.query_limit}
            """)
    database = dcc.Markdown(f"""
**Name:** {config.database.database_name}

**Port:** {config.database.port}

**Host:** {config.database.host}
        """)

    queue = dcc.Graph(figure=task_graph())
    return server, database, queue


def managers_table(status):
    socket = get_socket()

    managers = socket.get_managers(status=status)
    df = pd.DataFrame(managers["data"])

    df["uuid"] = [x[:6] for x in df["uuid"]]
    cols = ["cluster", "username", "tag", "submitted", "completed", "failures"]

    table = dash_table.DataTable(
        id='table',
        columns=[{
            "name": i.title(),
            "id": i
        } for i in cols],
        data=df[cols].to_dict('records'),
        filtering=True,
        sorting=True,
    )
    return table


groupby_items = dcc.Checklist(id="manager-overview-groupby",
                              options=[
                                  {
                                      'label': 'ACTIVE',
                                      'value': 'ACTIVE'
                                  },
                                  {
                                      'label': 'INACTIVE',
                                      'value': 'INACTIVE'
                                  },
                              ],
                              value=["ACTIVE"],
                              labelStyle={'display': 'inline-block'})

groupby_items2 = dcc.Checklist(id="manager-overview-groupby2",
                               options=[
                                   {
                                       'label': 'ACTIVE',
                                       'value': 'ACTIVE'
                                   },
                                   {
                                       'label': 'INACTIVE',
                                       'value': 'INACTIVE'
                                   },
                               ],
                               value=["ACTIVE"],
                               labelStyle={'display': 'inline-block'})

managers = dbc.Row([
    dbc.Col([
        coreui.AppCard([
            dbc.CardHeader("Manager Information"),
            dbc.CardBody([
                dbc.Col([
                    dbc.Row([groupby_items]),
                ]),
                dcc.Graph(id="manager-overview", style={"width": "100%"}),
            ])
        ])
    ],
            className="w-100")
],
                   className="w-100")

manager_list = dbc.Row([
    dbc.Col([
        coreui.AppCard([
            dbc.CardHeader("Manager List"),
            dbc.CardBody([
                dbc.Col([
                    dbc.Row([groupby_items2]),
                ]),
                dbc.Col(id="manager-list"),
            ])
        ])
    ],
            className="w-100")
],
                       className="w-100")


@app.callback(Output('manager-overview', 'figure'), [Input('manager-overview-groupby', 'value')])
def update_overview_graph(status):
    return manager_graph(status=status)


@app.callback(Output('manager-list', 'children'), [Input('manager-overview-groupby2', 'value')])
def update_overview_list(status):
    return list_managers(status=status)


### Build page return
layout = dbc.Row([top_row, managers, manager_list])

navbar = {
    "name": "Dashboard",
    "url": "/",
    "icon": "cui-speedometer icons",
    "badge": {
        "variant": "info",
        #        'text': 'NEW'
    },
}
data = {"layout": layout, "navitem": navbar, "id": "dashcard", "route": "/"}



"""Final migration synchronization from v0.11 to v0.12

Revision ID: 4b27843a188a
Revises: 159ba85908fd
Create Date: 2019-11-06 13:48:39.716633

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '4b27843a188a'
down_revision = '159ba85908fd'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('collection', 'group',
               existing_type=sa.VARCHAR(),
               type_=sa.String(length=100),
               existing_nullable=False)
    op.alter_column('molecule', 'geometry',
               existing_type=postgresql.BYTEA(),
               nullable=False)
    op.alter_column('molecule', 'symbols',
               existing_type=postgresql.BYTEA(),
               nullable=False)
    op.alter_column('task_queue', 'spec',
               existing_type=postgresql.BYTEA(),
               nullable=False)
    op.drop_constraint('task_queue_manager_fkey', 'task_queue', type_='foreignkey')
    op.create_foreign_key('task_queue_manager_fkey', 'task_queue', 'queue_manager', ['manager'], ['name'], ondelete='SET NULL')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('task_queue_manager_fkey', 'task_queue', type_='foreignkey')
    op.create_foreign_key('task_queue_manager_fkey', 'task_queue', 'queue_manager', ['manager'], ['name'])
    op.alter_column('task_queue', 'spec',
               existing_type=postgresql.BYTEA(),
               nullable=True)
    op.alter_column('molecule', 'symbols',
               existing_type=postgresql.BYTEA(),
               nullable=True)
    op.alter_column('molecule', 'geometry',
               existing_type=postgresql.BYTEA(),
               nullable=True)
    op.alter_column('collection', 'group',
               existing_type=sa.String(length=100),
               type_=sa.VARCHAR(),
               existing_nullable=False)
    # ### end Alembic commands ###
